"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __asyncValues = (this && this.__asyncValues) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator], i;
    return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
    function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
    function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.TaskQueue = void 0;
const Bucket_1 = require("./Bucket");
const PriorityQueue_1 = require("./PriorityQueue");
/**
 * Task Queue
 *
 * A Task Manager for handing varying priority tasks and scheduling.
 * `TaskQueue` instances provide two means for task scheduling.
 *
 * The first is a priority queue that indexes tasks based on a
 * priority specified by you:
 * ```
 * const TQ = new TaskQueue({
 *  priorities: 4,
 *  autoRun: false,
 *  taskSeparation: 5,
 *  mainThreadYieldTime: 5.
 * });
 *
 * const task1 = () => {};
 * const task2 = () => {};
 * const task3 = () => {};
 * const task4 = () => {};
 *
 * TQ.registerTask(task1, 4);
 * TQ.registerTask(task2, 3);
 * TQ.registerTask(task3, 2);
 * TQ.registerTask(task4, 1);
 *
 * const cancelFN = TQ.executeAll();
 * // The execution order => task4, task3, task2, task1
 * // 5ms will elapse between each task
 * ```
 *
 * `TaskQueue.registerTask()` returns a cancel function for easily
 * removing a task from the Queue
 *
 * The second means is the management and clean up of deferred tasks.
 * Deferring tasks through `setTimeout` is notorious for causing
 * memory leaks and unwanted side effects. This `TaskQueue` provides
 * management for scheduled tasks:
 * ```typescript
 * const TQ = new TaskQueue();
 *
 * const cancel = TQ.deferTask(() => {}, 1000); // Execute the task after 1000ms
 *
 * // cancel(); // cancel the task optionally
 * ```
 * `TaskQueue.deferTask()` also returns a cancel function for easily
 * removing a task from the Queue
 */
class TaskQueue {
    constructor(config = TaskQueue.defaultConfig) {
        this.autoRun = false;
        this.taskSeparation = 0;
        this.mainThreadYieldTime = 5;
        this.subscriptions = new Bucket_1.Bucket();
        this.internals = new Bucket_1.Bucket();
        this.deferredTasks = new Bucket_1.Bucket();
        const { priorities = 1, autoRun = false, taskSeparation = 0, mainThreadYieldTime = 5, } = config;
        this.autoRun = autoRun;
        this.taskSeparation = taskSeparation;
        this.mainThreadYieldTime = mainThreadYieldTime;
        this.tasks = new PriorityQueue_1.PriorityQueue(priorities);
    }
    /**
     * Register Task
     *
     * Registers a non-deferred task on the TaskQueue at
     * a specified priority level. Returns a cancel
     * function that removes the task from the queue
     */
    registerTask(task, priority = 1) {
        const ID = this.tasks.enqueue(task, priority - 1);
        if (this.autoRun && !this.internals.peek()) {
            this.executeAll();
        }
        return () => {
            this.tasks.delete(ID, priority - 1);
        };
    }
    /**
     * Defer Task
     *
     * Schedules the specified task after a delay. Returns
     * a cancel function that removes the task from the
     * queue
     */
    deferTask(task, delay) {
        const timer = setTimeout(() => {
            void task();
            this.deferredTasks.delete(ID);
        }, delay);
        const ID = this.deferredTasks.enqueue(timer);
        return () => {
            clearTimeout(timer);
            this.deferredTasks.delete(ID);
        };
    }
    /**
     * Execute All
     *
     * Executes all tasks registered using `TaskQueue.registerTask()`.
     * The task execution order will be based on the priority each task
     * was registered with. Returns a cancel function
     */
    executeAll(onComplete, taskSeparation = this.taskSeparation) {
        if (onComplete) {
            this.subscriptions.enqueue(onComplete);
        }
        if (this.internals.peek()) {
            return this.getCancelFN();
        }
        const cancelFN = this.cancellableExecution(this.tasks, taskSeparation);
        const jobID = this.internals.enqueue(cancelFN);
        const removeJob = () => {
            this.internals.delete(jobID);
        };
        this.subscriptions.enqueue(removeJob);
        return () => {
            cancelFN();
            removeJob();
        };
    }
    /**
     * Execute Tasks With Priority
     *
     * Executes all tasks registered using `TaskQueue.registerTask(task, N)`
     * with the provided priority. Returns a cancel function
     */
    executeTasksWithPriority(priority = 1, taskSeparation = 0, onComplete) {
        if (onComplete) {
            this.subscriptions.enqueue(onComplete);
        }
        if (this.internals.peek()) {
            return this.getCancelFN();
        }
        return this.cancellableExecution(this.tasks.getBucket(priority - 1), taskSeparation);
    }
    /**
     * Clear Pending Tasks
     *
     * Clears all pending tasks
     */
    clearPendingTasks() {
        this.tasks.clear();
        this.clearDeferredTasks();
    }
    /**
     * Clear Deferred Tasks
     *
     * Clears all pending tasks registered using `TaskQueue.deferTask()`
     */
    clearDeferredTasks() {
        while (this.deferredTasks.length) {
            const item = this.deferredTasks.dequeue();
            if (item !== undefined) {
                clearTimeout(item);
            }
        }
    }
    /**
     * Get Cancel Function
     *
     * Returns the cancel function for currently executing tasks
     * (if it exists). Returns undefined if no tasks are being
     * executed;
     */
    getCancelFN() {
        const activeTask = this.internals.peek();
        if (activeTask) {
            return activeTask[1];
        }
    }
    /**
     * Cancellable Execution
     *
     * Given Queue and a task separation time, schedules the
     * execution of each task present. Returns a cancel
     * function that pauses the execution of tasks
     */
    cancellableExecution(queue, taskSeparation = this.taskSeparation) {
        let cancelToken = false;
        const iterable = this.generator(queue, taskSeparation);
        const iterate = () => __awaiter(this, void 0, void 0, function* () {
            var _a, e_1, _b, _c;
            try {
                for (var _d = true, iterable_1 = __asyncValues(iterable), iterable_1_1; iterable_1_1 = yield iterable_1.next(), _a = iterable_1_1.done, !_a;) {
                    _c = iterable_1_1.value;
                    _d = false;
                    try {
                        const proceed = _c;
                        if (cancelToken || !proceed) {
                            break;
                        }
                        if (this.isInputPending()) {
                            yield this.yieldMainThread();
                            continue;
                        }
                        const task = queue.dequeue();
                        task && void task();
                    }
                    finally {
                        _d = true;
                    }
                }
            }
            catch (e_1_1) { e_1 = { error: e_1_1 }; }
            finally {
                try {
                    if (!_d && !_a && (_b = iterable_1.return)) yield _b.call(iterable_1);
                }
                finally { if (e_1) throw e_1.error; }
            }
            if (this.subscriptions.length) {
                return this.cancellableExecution(this.subscriptions, 0);
            }
        });
        void iterate();
        return () => {
            cancelToken = true;
        };
    }
    /**
     * Generator
     *
     * Returns an async iterable yielding a promise that resolves
     * after a specified duration. The duration specified is
     * used for dispersing prioritized along the call stack
     */
    *generator(queue, taskSeparation = 0) {
        while (!queue.isEmpty) {
            if (!taskSeparation) {
                yield true;
            }
            else {
                yield new Promise((resolve) => {
                    this.deferTask(() => {
                        resolve(true);
                    }, taskSeparation);
                });
            }
        }
    }
    isInputPending() {
        var _a, _b;
        if (typeof window !== "undefined" && typeof navigator !== "undefined") {
            // @ts-ignore
            return (_b = (_a = navigator === null || navigator === void 0 ? void 0 : navigator.scheduling) === null || _a === void 0 ? void 0 : _a.isInputPending()) !== null && _b !== void 0 ? _b : false;
        }
        return false;
    }
    yieldMainThread() {
        return new Promise((resolve) => {
            setTimeout(resolve, this.mainThreadYieldTime);
        });
    }
}
exports.TaskQueue = TaskQueue;
TaskQueue.defaultConfig = {
    priorities: 1,
    autoRun: false,
    taskSeparation: 0,
    mainThreadYieldTime: 5,
};
