/// <reference types="node" />
import { Bucket } from "./Bucket";
import { PriorityQueue } from "./PriorityQueue";
import type { CancelFN, ITaskQueue, Task } from "./types";
/**
 * Task Queue
 *
 * A Task Manager for handing varying priority tasks and scheduling.
 * `TaskQueue` instances provide two means for task scheduling.
 *
 * The first is a priority queue that indexes tasks based on a
 * priority specified by you:
 * ```
 * const TQ = new TaskQueue({
 *  priorities: 4,
 *  autoRun: false,
 *  taskSeparation: 5,
 *  mainThreadYieldTime: 5.
 * });
 *
 * const task1 = () => {};
 * const task2 = () => {};
 * const task3 = () => {};
 * const task4 = () => {};
 *
 * TQ.registerTask(task1, 4);
 * TQ.registerTask(task2, 3);
 * TQ.registerTask(task3, 2);
 * TQ.registerTask(task4, 1);
 *
 * const cancelFN = TQ.executeAll();
 * // The execution order => task4, task3, task2, task1
 * // 5ms will elapse between each task
 * ```
 *
 * `TaskQueue.registerTask()` returns a cancel function for easily
 * removing a task from the Queue
 *
 * The second means is the management and clean up of deferred tasks.
 * Deferring tasks through `setTimeout` is notorious for causing
 * memory leaks and unwanted side effects. This `TaskQueue` provides
 * management for scheduled tasks:
 * ```typescript
 * const TQ = new TaskQueue();
 *
 * const cancel = TQ.deferTask(() => {}, 1000); // Execute the task after 1000ms
 *
 * // cancel(); // cancel the task optionally
 * ```
 * `TaskQueue.deferTask()` also returns a cancel function for easily
 * removing a task from the Queue
 */
export declare class TaskQueue {
    autoRun: boolean;
    taskSeparation: number;
    mainThreadYieldTime: number;
    tasks: PriorityQueue<Task>;
    subscriptions: Bucket<Task>;
    internals: Bucket<CancelFN>;
    deferredTasks: Bucket<NodeJS.Timeout>;
    constructor(config?: ITaskQueue);
    static defaultConfig: ITaskQueue;
    /**
     * Register Task
     *
     * Registers a non-deferred task on the TaskQueue at
     * a specified priority level. Returns a cancel
     * function that removes the task from the queue
     */
    registerTask(task: Task, priority?: number): CancelFN;
    /**
     * Defer Task
     *
     * Schedules the specified task after a delay. Returns
     * a cancel function that removes the task from the
     * queue
     */
    deferTask(task: Task, delay: number): CancelFN;
    /**
     * Execute All
     *
     * Executes all tasks registered using `TaskQueue.registerTask()`.
     * The task execution order will be based on the priority each task
     * was registered with. Returns a cancel function
     */
    executeAll(onComplete?: Task, taskSeparation?: number): CancelFN;
    /**
     * Execute Tasks With Priority
     *
     * Executes all tasks registered using `TaskQueue.registerTask(task, N)`
     * with the provided priority. Returns a cancel function
     */
    executeTasksWithPriority(priority?: number, taskSeparation?: number, onComplete?: Task): CancelFN;
    /**
     * Clear Pending Tasks
     *
     * Clears all pending tasks
     */
    clearPendingTasks(): void;
    /**
     * Clear Deferred Tasks
     *
     * Clears all pending tasks registered using `TaskQueue.deferTask()`
     */
    clearDeferredTasks(): void;
    /**
     * Get Cancel Function
     *
     * Returns the cancel function for currently executing tasks
     * (if it exists). Returns undefined if no tasks are being
     * executed;
     */
    getCancelFN(): undefined | CancelFN;
    /**
     * Cancellable Execution
     *
     * Given Queue and a task separation time, schedules the
     * execution of each task present. Returns a cancel
     * function that pauses the execution of tasks
     */
    private cancellableExecution;
    /**
     * Generator
     *
     * Returns an async iterable yielding a promise that resolves
     * after a specified duration. The duration specified is
     * used for dispersing prioritized along the call stack
     */
    private generator;
    private isInputPending;
    private yieldMainThread;
}
